# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from typing import cast

import pytest

from elasticapm import get_client
from elasticapm.conf.constants import SPAN, TRANSACTION
from elasticapm.instrumentation.packages.psycopg2 import PGCursorProxy, extract_signature, get_destination_info
from elasticapm.utils import default_ports
from tests.fixtures import TempStoreClient

psycopg2 = pytest.importorskip("psycopg2")


try:
    from psycopg2 import compat

    is_cffi = True
except ImportError:
    is_cffi = False

try:
    import psycopg2.extensions
except ImportError:
    psycopg2.extensions = None


pytestmark = pytest.mark.psycopg2

has_postgres_configured = "POSTGRES_DB" in os.environ


def connect_kwargs():
    return {
        "database": os.environ.get("POSTGRES_DB", "elasticapm_test"),
        "user": os.environ.get("POSTGRES_USER", "postgres"),
        "password": os.environ.get("POSTGRES_PASSWORD", "postgres"),
        "host": os.environ.get("POSTGRES_HOST", None),
        "port": os.environ.get("POSTGRES_PORT", None),
    }


@pytest.fixture(scope="function")
def postgres_connection(request):
    conn = psycopg2.connect(**connect_kwargs())
    cursor = conn.cursor()
    cursor.execute(
        "CREATE TABLE test(id int, name VARCHAR(5) NOT NULL);"
        "INSERT INTO test VALUES (1, 'one'), (2, 'two'), (3, 'three');"
    )

    yield conn

    # cleanup
    cursor.execute("ROLLBACK")


def test_insert():
    sql_statement = """INSERT INTO mytable (id, name) VALUE ('2323', 'Ron')"""
    actual = extract_signature(sql_statement)

    assert "INSERT INTO mytable" == actual


def test_update_with_quotes():
    sql_statement = """UPDATE "my table" set name='Ron' WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "UPDATE my table" == actual


def test_update():
    sql_statement = """update mytable set name = 'Ron where id = 'a'"""
    actual = extract_signature(sql_statement)

    assert "UPDATE mytable" == actual


def test_delete_simple():
    sql_statement = 'DELETE FROM "mytable"'
    actual = extract_signature(sql_statement)

    assert "DELETE FROM mytable" == actual


def test_delete():
    sql_statement = """DELETE FROM "my table" WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "DELETE FROM my table" == actual


def test_select_simple():
    sql_statement = """SELECT id, name FROM my_table WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM my_table" == actual


def test_select_with_entity_quotes():
    sql_statement = """SELECT id, name FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM mytable" == actual


def test_select_with_difficult_values():
    sql_statement = """SELECT id, 'some name' + '" from Denmark' FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM mytable" == actual


def test_select_with_dollar_quotes():
    sql_statement = """SELECT id, $$some single doubles ' $$ + '" from Denmark' FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM mytable" == actual


def test_select_with_invalid_dollar_quotes():
    sql_statement = """SELECT id, $fish$some single doubles ' $$ + '" from Denmark' FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM" == actual


def test_select_with_dollar_quotes_custom_token():
    sql_statement = """SELECT id, $token $FROM $ FROM $ FROM single doubles ' $token $ + '" from Denmark' FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM mytable" == actual


def test_select_with_difficult_table_name():
    sql_statement = """SELECT id FROM "myta\n-æøåble" WHERE id = 2323"""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM myta\n-æøåble" == actual


def test_select_subselect():
    sql_statement = """SELECT id, name FROM (
            SELECT id, 'not a FROM ''value' FROM mytable WHERE id = 2323
    ) LIMIT 20"""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM mytable" == actual


def test_select_subselect_with_alias():
    sql_statement = """
    SELECT count(*)
    FROM (
        SELECT count(id) AS some_alias, some_column
        FROM mytable
        GROUP BY some_colun
        HAVING count(id) > 1
    ) AS foo
    """
    actual = extract_signature(sql_statement)

    assert "SELECT FROM mytable" == actual


def test_select_with_multiple_tables():
    sql_statement = """SELECT count(table2.id)
        FROM table1, table2, table2
        WHERE table2.id = table1.table2_id
    """
    actual = extract_signature(sql_statement)
    assert "SELECT FROM table1" == actual


def test_select_with_invalid_subselect():
    sql_statement = "SELECT id FROM (SELECT * " ""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM" == actual


def test_select_with_invalid_literal():
    sql_statement = "SELECT 'neverending literal FROM (SELECT * FROM ..." ""
    actual = extract_signature(sql_statement)

    assert "SELECT FROM" == actual


def test_savepoint():
    sql_statement = """SAVEPOINT x_asd1234"""
    actual = extract_signature(sql_statement)

    assert "SAVEPOINT" == actual


def test_begin():
    sql_statement = """BEGIN"""
    actual = extract_signature(sql_statement)

    assert "BEGIN" == actual


def test_create_index_with_name():
    sql_statement = """CREATE INDEX myindex ON mytable"""
    actual = extract_signature(sql_statement)

    assert "CREATE INDEX" == actual


def test_create_index_without_name():
    sql_statement = """CREATE INDEX ON mytable"""
    actual = extract_signature(sql_statement)

    assert "CREATE INDEX" == actual


def test_drop_table():
    sql_statement = """DROP TABLE mytable"""
    actual = extract_signature(sql_statement)

    assert "DROP TABLE" == actual


def test_multi_statement_sql():
    sql_statement = """CREATE TABLE mytable; SELECT * FROM mytable; DROP TABLE mytable"""
    actual = extract_signature(sql_statement)

    assert "CREATE TABLE" == actual


def test_fully_qualified_table_name():
    sql_statement = """SELECT a.b FROM db.schema.mytable as a;"""
    actual = extract_signature(sql_statement)
    assert "SELECT FROM db.schema.mytable" == actual


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_cursor_execute_signature(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    cursor.execute(query="SELECT 1", vars=None)
    row = cursor.fetchone()

    assert row


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_cursor_executemany_signature(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    res = cursor.executemany(
        query="INSERT INTO test VALUES (%s, %s)",
        vars_list=((4, "four"),),
    )
    assert res is None


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_destination(instrument, postgres_connection, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    cursor = postgres_connection.cursor()
    cursor.execute("SELECT 1")
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[TRANSACTION][0]
    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert span["context"]["destination"] == {
        "address": os.environ.get("POSTGRES_HOST", None),
        "port": default_ports["postgresql"],
        "service": {"name": "", "resource": "postgresql/elasticapm_test", "type": ""},
    }


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg2_register_type(instrument, postgres_connection, elasticapm_client):
    import psycopg2.extras

    elasticapm_client.begin_transaction("web.django")
    new_type = psycopg2.extras.register_uuid(None, postgres_connection)
    elasticapm_client.end_transaction(None, "test-transaction")

    assert new_type is not None


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg2_register_json(instrument, postgres_connection, elasticapm_client):
    # register_json bypasses register_type, so we have to test unwrapping
    # separately
    import psycopg2.extras

    elasticapm_client.begin_transaction("web.django")
    # as arg
    new_type = psycopg2.extras.register_json(postgres_connection, loads=lambda x: x)
    assert new_type is not None
    # as kwarg
    new_type = psycopg2.extras.register_json(conn_or_curs=postgres_connection, loads=lambda x: x)
    assert new_type is not None
    elasticapm_client.end_transaction(None, "test-transaction")


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
@pytest.mark.skipif(
    not hasattr(psycopg2.extensions, "quote_ident"), reason="psycopg2 driver doesn't have quote_ident extension"
)
def test_psycopg2_quote_ident(instrument, postgres_connection, elasticapm_client):
    elasticapm_client.begin_transaction("web.django")
    ident = psycopg2.extensions.quote_ident("x'x", postgres_connection)
    elasticapm_client.end_transaction(None, "test-transaction")

    assert ident == '"x\'x"'


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
@pytest.mark.skipif(
    not hasattr(psycopg2.extensions, "encrypt_password"),
    reason="psycopg2 driver doesn't have encrypt_password extension",
)
@pytest.mark.skipif(
    hasattr(psycopg2, "__libpq_version__") and psycopg2.__libpq_version__ < 100000,
    reason="test code requires libpq >= 10",
)
def test_psycopg2_encrypt_password(instrument, postgres_connection, elasticapm_client):
    elasticapm_client.begin_transaction("web.django")
    pw1 = psycopg2.extensions.encrypt_password("user", "password", postgres_connection)
    pw2 = psycopg2.extensions.encrypt_password("user", "password", postgres_connection, None)
    pw3 = psycopg2.extensions.encrypt_password("user", "password", postgres_connection, algorithm=None)
    pw4 = psycopg2.extensions.encrypt_password("user", "password", scope=postgres_connection, algorithm=None)
    elasticapm_client.end_transaction(None, "test-transaction")

    assert pw1.startswith("md5") and (pw1 == pw2 == pw3 == pw4)


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg2_tracing_outside_of_elasticapm_transaction(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    # check that the cursor is a proxy, even though we're not in an elasticapm
    # transaction
    assert isinstance(cursor, PGCursorProxy)
    cursor.execute("SELECT 1")
    transactions = elasticapm_client.events[TRANSACTION]
    assert not transactions


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg2_select_LIKE(instrument, postgres_connection, elasticapm_client):
    """
    Check that we pass queries with %-notation but without parameters
    properly to the dbapi backend
    """
    cursor = postgres_connection.cursor()
    query = "SELECT * FROM test WHERE name LIKE 't%'"

    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute(query)
        cursor.fetchall()
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        # make sure we've cleared out the spans for the other tests.
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])
        span = spans[0]
        assert span["name"] == "SELECT FROM test"
        assert span["type"] == "db"
        assert span["subtype"] == "postgresql"
        assert span["action"] == "query"
        assert "db" in span["context"]
        assert span["context"]["db"]["instance"] == "elasticapm_test"
        assert span["context"]["db"]["type"] == "sql"
        assert span["context"]["db"]["statement"] == query
        assert span["context"]["service"]["target"]["type"] == "postgresql"
        assert span["context"]["service"]["target"]["name"] == "elasticapm_test"


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
@pytest.mark.skipif(is_cffi, reason="psycopg2cffi does not have the sql module")
def test_psycopg2_composable_query_works(instrument, postgres_connection, elasticapm_client):
    """
    Check that we parse queries that are psycopg2.sql.Composable correctly
    """
    from psycopg2 import sql

    cursor = postgres_connection.cursor()
    query = sql.SQL("SELECT * FROM {table} WHERE {row} LIKE 't%' ORDER BY {row} DESC").format(
        table=sql.Identifier("test"), row=sql.Identifier("name")
    )
    baked_query = query.as_string(cursor.__wrapped__)
    result = None
    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute(query)
        result = cursor.fetchall()
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        # make sure we've cleared out the spans for the other tests.
        assert [(2, "two"), (3, "three")] == result
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])
        span = spans[0]
        assert span["name"] == "SELECT FROM test"
        assert "db" in span["context"]
        assert span["context"]["db"]["instance"] == "elasticapm_test"
        assert span["context"]["db"]["type"] == "sql"
        assert span["context"]["db"]["statement"] == baked_query


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg2_call_stored_procedure(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    cursor.execute("""
        CREATE OR REPLACE FUNCTION squareme(me INT)
        RETURNS INTEGER
        LANGUAGE SQL
        AS $$
            SELECT me*me;
        $$;
        """)
    elasticapm_client.begin_transaction("test")
    cursor.callproc("squareme", [2])
    result = cursor.fetchall()
    assert result[0][0] == 4
    elasticapm_client.end_transaction("test", "OK")
    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]
    assert span["name"] == "squareme()"
    assert span["action"] == "exec"


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg_context_manager(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    with psycopg2.connect(**connect_kwargs()) as conn:
        with conn.cursor() as curs:
            curs.execute("SELECT 1;")
            curs.fetchall()
    elasticapm_client.end_transaction("test", "OK")
    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert len(spans) == 2
    assert spans[0]["subtype"] == "postgresql"
    assert spans[0]["action"] == "connect"
    assert spans[0]["context"]["service"]["target"]["type"] == "postgresql"
    assert spans[0]["context"]["service"]["target"]["name"] == "elasticapm_test"

    assert spans[1]["subtype"] == "postgresql"
    assert spans[1]["action"] == "query"


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg2_rows_affected(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute("INSERT INTO test VALUES (4, 'four')")
        cursor.execute("SELECT * FROM test")
        cursor.execute("UPDATE test SET name = 'five' WHERE  id = 4")
        cursor.execute("DELETE FROM test WHERE  id = 4")
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])

        assert spans[0]["name"] == "INSERT INTO test"
        assert spans[0]["context"]["db"]["rows_affected"] == 1

        assert spans[1]["name"] == "SELECT FROM test"
        assert "rows_affected" not in spans[1]["context"]["db"]

        assert spans[2]["name"] == "UPDATE test"
        assert spans[2]["context"]["db"]["rows_affected"] == 1

        assert spans[3]["name"] == "DELETE FROM test"
        assert spans[3]["context"]["db"]["rows_affected"] == 1


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
@pytest.mark.skipif(is_cffi, reason="psycopg2cffi doesn't have execute_values")
def test_psycopg2_execute_values(instrument, postgres_connection, elasticapm_client):
    from psycopg2.extras import execute_values

    cursor = postgres_connection.cursor()
    try:
        elasticapm_client.begin_transaction("web.django")
        query = "INSERT INTO test VALUES %s"
        data = tuple((i, "xxxxx") for i in range(999))
        # this creates a long (~14000 characters) query, encoded as byte string.
        # This tests that we shorten already-encoded strings.
        execute_values(cursor, query, data, page_size=1000)
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])
        assert spans[0]["name"] == "INSERT INTO test"
        assert len(spans[0]["context"]["db"]["statement"]) == 10000, spans[0]["context"]["db"]["statement"]


@pytest.mark.integrationtest
def test_psycopg2_connection(instrument, elasticapm_transaction, postgres_connection):
    # elastciapm_client.events is only available on `TempStoreClient`, this keeps the type checkers happy
    elasticapm_client = cast(TempStoreClient, get_client())
    elasticapm_client.end_transaction("test", "success")
    span = elasticapm_client.events[SPAN][0]
    host = os.environ.get("POSTGRES_HOST", "localhost")
    assert span["name"] == f"psycopg2.connect {host}:5432"
    assert span["action"] == "connect"


@pytest.mark.parametrize(
    "host_in,port_in,expected_host_out,expected_port_out",
    (
        (None, None, "localhost", 5432),
        (None, 5432, "localhost", 5432),
        ("localhost", "5432", "localhost", 5432),
        ("foo.bar", "5432", "foo.bar", 5432),
        ("localhost,foo.bar", "5432,1234", "localhost", 5432),  # multiple hosts
    ),
)
def test_get_destination(host_in, port_in, expected_host_out, expected_port_out):
    host, port = get_destination_info(host_in, port_in)
    assert host == expected_host_out
    assert port == expected_port_out



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import types as python_types
import typing
import urllib.parse
from typing import Optional, Union

from opentelemetry.context import Context
from opentelemetry.sdk import trace as oteltrace
from opentelemetry.trace import SpanKind
from opentelemetry.trace.propagation import _SPAN_KEY
from opentelemetry.trace.span import SpanContext, TraceFlags, TraceState
from opentelemetry.trace.status import Status, StatusCode
from opentelemetry.util import types

import elasticapm
import elasticapm.conf.constants as constants
import elasticapm.traces


class Span(oteltrace.Span):
    """
    This is a wrapper around an Elastic APM Span/Transaction object, to match the
    otel Span API
    """

    def __init__(
        self,
        name: str,
        elastic_span: elasticapm.traces.BaseSpan,
        set_status_on_exception: Optional[bool] = None,
        client: Optional[elasticapm.Client] = None,
    ) -> None:
        self.elastic_span = elastic_span
        self.otel_context = Context({_SPAN_KEY: self})
        elastic_span.otel_wrapper = self
        self.set_status_on_exception = set_status_on_exception
        self.client = client if client else elasticapm.get_client()
        self._name = name

    def end(self, end_time: Optional[int] = None) -> None:
        """Sets the current time as the span's end time.
        The span's end time is the wall time at which the operation finished.
        Only the first call to `end` should modify the span, and
        implementations are free to ignore or raise on further calls.
        """
        is_transaction = isinstance(self.elastic_span, elasticapm.traces.Transaction)
        if self.elastic_span.ended_time:
            # Already ended
            return
        self._set_types()
        if end_time:
            if is_transaction:
                self.client.end_transaction(
                    name=self._name,
                    result=self.elastic_span.outcome or "OK",
                    duration=end_time - self.elastic_span.timestamp,
                )
            else:
                self.elastic_span.end(duration=end_time - self.elastic_span.timestamp)
        else:
            if is_transaction:
                self.client.end_transaction(name=self._name, result=self.elastic_span.outcome or "OK")
            else:
                self.elastic_span.end()

    def get_span_context(self) -> "SpanContext":
        """Gets the span's SpanContext.
        Get an immutable, serializable identifier for this span that can be
        used to create new child spans.
        Returns:
            A :class:`opentelemetry.trace.SpanContext` with a copy of this span's immutable state.
        """
        return SpanContext(
            trace_id=int(self.elastic_span.transaction.trace_parent.trace_id, base=16),
            span_id=int(self.elastic_span.id, base=16),
            is_remote=False,
            trace_flags=TraceFlags(
                TraceFlags.SAMPLED if self.elastic_span.transaction.is_sampled else TraceFlags.DEFAULT
            ),
            trace_state=TraceState(list(self.elastic_span.transaction.trace_parent.tracestate_dict.items())),
        )

    def set_attributes(self, attributes: typing.Dict[str, types.AttributeValue]) -> None:
        """Sets Attributes.
        Sets Attributes with the key and value passed as arguments dict.
        Note: The behavior of `None` value attributes is undefined, and hence strongly discouraged.
        """
        if not attributes:
            return
        for key, value in attributes.items():
            self.set_attribute(key, value)

    def set_attribute(self, key: str, value: types.AttributeValue) -> None:
        """Sets an Attribute.
        Sets a single Attribute with the key and value passed as arguments.
        Note: The behavior of `None` value attributes is undefined, and hence strongly discouraged.
        """
        if "otel_attributes" not in self.elastic_span.context:
            self.elastic_span.context["otel_attributes"] = {}
        self.elastic_span.context["otel_attributes"][key] = value

    def add_event(
        self,
        name: str,
        attributes: types.Attributes = None,
        timestamp: Optional[int] = None,
    ) -> None:
        """Adds an `Event`.
        Adds a single `Event` with the name and, optionally, a timestamp and
        attributes passed as arguments. Implementations should generate a
        timestamp if the `timestamp` argument is omitted.
        """
        raise NotImplementedError("Events are not implemented in the otel bridge at this time")

    def update_name(self, name: str) -> None:
        """Updates the `Span` name.
        This will override the name provided via :func:`opentelemetry.trace.Tracer.start_span`.
        Upon this update, any sampling behavior based on Span name will depend
        on the implementation.
        """
        self.elastic_span.name = name

    def is_recording(self) -> bool:
        """Returns whether this span will be recorded.
        Returns true if this Span is active and recording information like
        events with the add_event operation and attributes using set_attribute.
        """
        return self.elastic_span.transaction.is_sampled and not self.elastic_span.ended_time

    def set_status(self, status: Union[Status, StatusCode], description: Optional[str] = None) -> None:
        """Sets the Status of the Span. If used, this will override the default
        Span status.
        """
        # Handle both Status objects and StatusCode enums
        if isinstance(status, Status):
            status_code = status.status_code
        else:
            status_code = status

        if status_code == StatusCode.ERROR:
            self.elastic_span.outcome = constants.OUTCOME.FAILURE
        elif status_code == StatusCode.OK:
            self.elastic_span.outcome = constants.OUTCOME.SUCCESS
        else:
            self.elastic_span.outcome = constants.OUTCOME.UNKNOWN

    def record_exception(
        self,
        exception: Exception,
        attributes: types.Attributes = None,
        timestamp: Optional[int] = None,
        escaped: bool = False,
    ) -> None:
        """Records an exception as a span event."""
        client = elasticapm.get_client()
        client.capture_exception(
            exc_info=(type(exception), exception, exception.__traceback__),
            date=datetime.datetime.fromtimestamp(timestamp),
            context={"otel_attributes": attributes} if attributes else None,
            handled=escaped,
        )

    def __enter__(self) -> "Span":
        """Invoked when `Span` is used as a context manager.
        Returns the `Span` itself.
        """
        return self

    def __exit__(
        self,
        exc_type: Optional[typing.Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[python_types.TracebackType],
    ) -> None:
        """Ends context manager and calls `end` on the `Span`."""

        self.end()

    def _set_types(self) -> None:
        """
        Set the types and subtypes for the underlying Elastic transaction/span
        """
        if isinstance(self.elastic_span, elasticapm.traces.Transaction):
            attributes = self.elastic_span.context.get("otel_attributes", {})
            span_kind = self.elastic_span.context["otel_spankind"]
            is_rpc = bool(attributes.get("rpc.system"))
            is_http = bool(attributes.get("http.url")) or bool(attributes.get("http.scheme"))
            is_messaging = bool(attributes.get("messaging.system"))
            if span_kind == SpanKind.SERVER.name and (is_rpc or is_http):
                transaction_type = "request"
            elif span_kind == SpanKind.CONSUMER.name and is_messaging:
                transaction_type = "messaging"
            else:
                transaction_type = "unknown"
            self.elastic_span.transaction_type = transaction_type
        else:
            attributes = self.elastic_span.context.get("otel_attributes", {})
            span_type = None
            span_subtype = None
            resource = None

            def http_port_from_scheme(scheme: str):
                if scheme == "http":
                    return 80
                elif scheme == "https":
                    return 443
                return None

            def parse_net_name(url: str):
                u = urllib.parse.urlparse(url)
                if u.port:
                    return u.netloc
                else:
                    port = http_port_from_scheme(u.scheme)
                    return u.netloc if not port else "{}:{}".format(u.netloc, port)

            net_port = attributes.get("net.peer.port", -1)
            net_name = net_peer = attributes.get("net.peer.name", attributes.get("net.peer.ip", ""))

            if net_name and (net_port > 0):
                net_name = f"{net_name}:{net_port}"

            if attributes.get("db.system"):
                span_type = "db"
                span_subtype = attributes.get("db.system")
                resource = net_name or span_subtype
                if attributes.get("db.name"):
                    resource = "{}/{}".format(resource, attributes.get("db.name"))
            elif attributes.get("messaging.system"):
                span_type = "messaging"
                span_subtype = attributes.get("messaging.system")
                if not net_name and attributes.get("messaging.url"):
                    net_name = parse_net_name(attributes.get("messaging.url"))
                resource = net_name or span_subtype
                if attributes.get("messaging.destination"):
                    resource = "{}/{}".format(resource, attributes.get("messaging.destination"))
            elif attributes.get("rpc.system"):
                span_type = "external"
                span_subtype = attributes.get("rpc.system")
                resource = net_name or span_subtype
                if attributes.get("rpc.service"):
                    resource = "{}/{}".format(resource, attributes.get("rpc.service"))
            elif attributes.get("http.url") or attributes.get("http.scheme"):
                span_type = "external"
                span_subtype = "http"
                http_host = attributes.get("http.host", net_peer)
                if http_host:
                    if net_port < 0:
                        net_port = http_port_from_scheme(attributes.get("http.scheme"))
                    resource = http_host if net_port < 0 else f"{http_host}:{net_port}"
                elif attributes.get("http.url"):
                    resource = parse_net_name(attributes["http.url"])

            if not span_type:
                span_kind = self.elastic_span.context["otel_spankind"]
                if span_kind == SpanKind.INTERNAL.name:
                    span_type = "app"
                    span_subtype = "internal"
                else:
                    span_type = "unknown"
            self.elastic_span.type = span_type
            self.elastic_span.subtype = span_subtype
            if resource:
                if "destination" not in self.elastic_span.context:
                    self.elastic_span.context["destination"] = {"service": {"resource": resource}}
                elif "service" not in self.elastic_span.context["destination"]:
                    self.elastic_span.context["destination"]["service"] = {"resource": resource}
                else:
                    self.elastic_span.context["destination"]["service"]["resource"] = resource



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

async_asgi_testclient = pytest.importorskip("async_asgi_testclient")  # isort:skip

from elasticapm.conf import constants
from elasticapm.contrib.asgi import ASGITracingMiddleware
from tests.contrib.asgi.app import app

pytestmark = pytest.mark.asgi


@pytest.fixture(scope="function")
def instrumented_app(elasticapm_client):
    return ASGITracingMiddleware(app)


@pytest.mark.asyncio
async def test_transaction_span_success(instrumented_app, elasticapm_client):
    async with async_asgi_testclient.TestClient(instrumented_app) as client:
        resp = await client.get("/")
        assert resp.status_code == 200
        assert resp.text == "OK"

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    assert len(elasticapm_client.events[constants.SPAN]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    span = elasticapm_client.events[constants.SPAN][0]
    assert transaction["name"] == "GET unknown route"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"
    assert transaction["context"]["request"]["url"]["full"] == "/"
    assert transaction["context"]["response"]["status_code"] == 200

    assert span["name"] == "sleep"
    assert span["outcome"] == "success"
    assert span["sync"] == False


@pytest.mark.asyncio
async def test_transaction_span_failure(instrumented_app, elasticapm_client):
    async with async_asgi_testclient.TestClient(instrumented_app) as client:
        resp = await client.get("/500")
        assert resp.status_code == 500
        assert resp.text == "KO"

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    assert len(elasticapm_client.events[constants.SPAN]) == 0
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["name"] == "GET unknown route"
    assert transaction["result"] == "HTTP 5xx"
    assert transaction["outcome"] == "failure"
    assert transaction["context"]["request"]["url"]["full"] == "/500"
    assert transaction["context"]["response"]["status_code"] == 500


@pytest.mark.asyncio
async def test_transaction_traceparent(instrumented_app, elasticapm_client):
    async with async_asgi_testclient.TestClient(instrumented_app) as client:
        resp = await client.get("/", headers={"traceparent": "00-12345678901234567890123456789012-1234567890123456-01"})
        assert resp.status_code == 200
        assert resp.text == "OK"

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    assert len(elasticapm_client.events[constants.SPAN]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    span = elasticapm_client.events[constants.SPAN][0]
    assert transaction["name"] == "GET unknown route"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"
    assert transaction["context"]["request"]["url"]["full"] == "/"
    assert transaction["context"]["response"]["status_code"] == 200

    assert transaction["trace_id"] == "12345678901234567890123456789012"

    assert span["name"] == "sleep"
    assert span["outcome"] == "success"
    assert span["sync"] == False


@pytest.mark.asyncio
async def test_transaction_ignore_url(instrumented_app, elasticapm_client):
    elasticapm_client.config.update("1", transaction_ignore_urls="/foo*")

    async with async_asgi_testclient.TestClient(instrumented_app) as client:
        resp = await client.get("/foo")
        assert resp.status_code == 200
        assert resp.text == "foo"

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 0


@pytest.mark.asyncio
async def test_transaction_headers(instrumented_app, elasticapm_client):
    elasticapm_client.config.update("1", capture_headers="true")

    async with async_asgi_testclient.TestClient(instrumented_app) as client:
        resp = await client.get("/foo", headers={"baz": "bazzinga"})
        assert resp.status_code == 200
        assert resp.text == "foo"

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["context"]["request"]["headers"]["baz"] == "bazzinga"
    assert transaction["context"]["response"]["headers"]["foo"] == "bar"

    elasticapm_client.config.update("1", capture_headers="false")

    async with async_asgi_testclient.TestClient(instrumented_app) as client:
        resp = await client.get("/foo", headers={"baz": "bazzinga"})
        assert resp.status_code == 200
        assert resp.text == "foo"

    transaction = elasticapm_client.events[constants.TRANSACTION][1]
    assert "headers" not in transaction["context"]["request"]
    assert "headers" not in transaction["context"]["response"]


@pytest.mark.asyncio
async def test_transaction_body(instrumented_app, elasticapm_client):
    elasticapm_client.config.update("1", capture_body="transactions")

    async with async_asgi_testclient.TestClient(instrumented_app) as client:
        resp = await client.post("/", data="foo")
        assert resp.status_code == 200
        assert resp.text == "OK"

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert "body" in transaction["context"]["request"]
    assert transaction["context"]["request"]["body"] == "foo"


# for some reason, exceptions don't seem to bubble up to our middleware with the two ASGI
# frameworks I tested (Quart, Sanic)

# @pytest.mark.asyncio
# async def test_transaction_exception(instrumented_app, elasticapm_client):
#     async with async_asgi_testclient.TestClient(instrumented_app) as client:
#         resp = await client.get("/boom")
#         assert resp.status_code == 500
#
#     assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
#
#     transaction = elasticapm_client.events[constants.TRANSACTION][0]
#     assert transaction["name"] == "GET unknown route"
#     assert transaction["result"] == "HTTP 5xx"
#     # assert transaction["outcome"] == "failure"
#
#     assert len(elasticapm_client.events[constants.ERROR]) == 1
#     error = elasticapm_client.events[constants.ERROR][0]
#     pass



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import pytest

from elasticapm.conf.constants import TRANSACTION

pyodbc = pytest.importorskip("pyodbc")

pytestmark = [pytest.mark.pyodbc]


if "POSTGRES_DB" not in os.environ:
    pytestmark.append(pytest.mark.skip("Skipping pyodbc tests, no POSTGRES_DB environment variable set"))


@pytest.fixture(scope="function")
def pyodbc_postgres_connection(request):
    conn_str = ("DRIVER={PostgreSQL Unicode};" "DATABASE=%s;" "UID=%s;" "PASSWORD=%s;" "SERVER=%s;" "PORT=%s;") % (
        os.environ.get("POSTGRES_DB", "elasticapm_test"),
        os.environ.get("POSTGRES_USER", "postgres"),
        os.environ.get("POSTGRES_PASSWORD", "postgres"),
        os.environ.get("POSTGRES_HOST", None),
        os.environ.get("POSTGRES_PORT", "5432"),
    )
    conn = pyodbc.connect(conn_str)
    cursor = conn.cursor()
    cursor.execute(
        "CREATE TABLE test(id int, name VARCHAR(5) NOT NULL);"
        "INSERT INTO test VALUES (1, 'one'), (2, 'two'), (3, 'three');"
    )

    yield conn

    # cleanup
    cursor.execute("ROLLBACK")


@pytest.mark.integrationtest
def test_pyodbc_select(instrument, pyodbc_postgres_connection, elasticapm_client):
    cursor = pyodbc_postgres_connection.cursor()
    query = "SELECT * FROM test WHERE name LIKE 't%'"

    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute(query)
        cursor.fetchall()
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])
        span = spans[0]
        assert span["name"] == "SELECT FROM test"
        assert span["type"] == "db"
        assert span["subtype"] == "pyodbc"
        assert span["action"] == "query"
        assert "db" in span["context"]
        assert span["context"]["db"]["instance"] == "elasticapm_test"
        assert span["context"]["db"]["type"] == "sql"
        assert span["context"]["db"]["statement"] == query
        assert span["context"]["service"]["target"]["type"] == "pyodbc"
        assert span["context"]["service"]["target"]["name"] == "elasticapm_test"


@pytest.mark.integrationtest
def test_psycopg2_rows_affected(instrument, pyodbc_postgres_connection, elasticapm_client):
    cursor = pyodbc_postgres_connection.cursor()
    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute("INSERT INTO test VALUES (4, 'four')")
        cursor.execute("SELECT * FROM test")
        cursor.execute("UPDATE test SET name = 'five' WHERE  id = 4")
        cursor.execute("DELETE FROM test WHERE  id = 4")
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])

        assert spans[0]["name"] == "INSERT INTO test"
        assert spans[0]["context"]["db"]["rows_affected"] == 1

        assert spans[1]["name"] == "SELECT FROM test"
        assert "rows_affected" not in spans[1]["context"]["db"]

        assert spans[2]["name"] == "UPDATE test"
        assert spans[2]["context"]["db"]["rows_affected"] == 1

        assert spans[3]["name"] == "DELETE FROM test"
        assert spans[3]["context"]["db"]["rows_affected"] == 1



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import pytest

from elasticapm.conf.constants import TRANSACTION
from elasticapm.instrumentation.packages.mysql import extract_signature

mysqldb = pytest.importorskip("MySQLdb")


pytestmark = [pytest.mark.mysqlclient]

if "MYSQL_HOST" not in os.environ:
    pytestmark.append(pytest.mark.skip("Skipping mysqlclient tests, no MYSQL_HOST environment variable set"))


@pytest.fixture(scope="function")
def mysqlclient_connection(request):
    conn = mysqldb.connect(
        host=os.environ.get("MYSQL_HOST", "localhost"),
        user=os.environ.get("MYSQL_USER", "eapm"),
        password=os.environ.get("MYSQL_PASSWORD", "Very(!)Secure"),
        database=os.environ.get("MYSQL_DATABASE", "eapm_tests"),
    )
    cursor = conn.cursor()
    cursor.execute("CREATE TABLE `test` (`id` INT, `name` VARCHAR(5))")
    cursor.execute("INSERT INTO `test` (`id`, `name`) VALUES (1, 'one'), (2, 'two'), (3, 'three')")
    row = cursor.fetchone()

    yield conn

    cursor.execute("DROP TABLE `test`")


@pytest.mark.integrationtest
def test_mysql_connector_select(instrument, mysqlclient_connection, elasticapm_client):
    cursor = mysqlclient_connection.cursor()
    query = "SELECT * FROM test WHERE name LIKE 't%' ORDER BY id"

    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute(query)
        assert cursor.fetchall() == ((2, "two"), (3, "three"))
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])
        span = spans[0]
        assert span["name"] == "SELECT FROM test"
        assert span["type"] == "db"
        assert span["subtype"] == "mysql"
        assert span["action"] == "query"
        assert "db" in span["context"]
        assert span["context"]["db"]["instance"] == "eapm_tests"
        assert span["context"]["db"]["type"] == "sql"
        assert span["context"]["db"]["statement"] == query
        assert span["context"]["destination"] == {
            "address": "mysql",
            "port": 3306,
            "service": {"name": "", "resource": "mysql/eapm_tests", "type": ""},
        }
        assert span["context"]["service"]["target"]["type"] == "mysql"
        assert span["context"]["service"]["target"]["name"] == "eapm_tests"



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import logging
import os

import mock
import pytest

import elasticapm
from elasticapm import Client, processors
from elasticapm.conf.constants import BASE_SANITIZE_FIELD_NAMES_UNPROCESSED, ERROR, SPAN, TRANSACTION
from tests.utils import assert_any_record_contains


@pytest.fixture()
def http_test_data():
    return {
        "context": {
            "request": {
                "body": "foo=bar&password=123456&secret=abc&cc=1234567890098765&custom_field=123",
                "env": {
                    "foo": "bar",
                    "password": "hello",
                    "secret": "hello",
                    "custom_env": "bye",
                },
                "headers": {
                    "foo": "bar",
                    "password": "hello",
                    "secret": "hello",
                    "authorization": "bearer xyz",
                    "some-header": "some-secret-value",
                    "cookie": "foo=bar; baz=foo",
                    "Ms-Client-Principal-Id": "foo",
                },
                "cookies": {
                    "foo": "bar",
                    "password": "topsecret",
                    "secret": "topsecret",
                    "sessionid": "123",
                    "custom-sensitive-cookie": "123",
                },
                "url": {
                    "full": "http://example.com/bla?foo=bar&password=123456&secret=abc&cc=1234567890098765&custom_field=123",
                    "search": "foo=bar&password=123456&secret=abc&cc=1234567890098765&custom_field=123",
                },
            },
            "response": {
                "status_code": "200",
                "headers": {
                    "foo": "bar",
                    "password": "hello",
                    "secret": "hello",
                    "authorization": "bearer xyz",
                    "some-header": "some-secret-value",
                    "cookie": "foo=bar; baz=foo",
                    "Ms-Client-Principal-Id": "foo",
                },
            },
        }
    }


@pytest.mark.parametrize(
    "elasticapm_client, custom_field",
    [
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["sensitive-stacktrace-val"]},
            processors.MASK,
        ),
        ({}, "bye"),
    ],
    indirect=["elasticapm_client"],
)
def test_stacktrace(elasticapm_client, custom_field):
    data = {
        "exception": {
            "stacktrace": [
                {
                    "vars": {
                        "foo": "bar",
                        "password": "hello",
                        "secret": "hello",
                        "sensitive-stacktrace-val": "bye",
                    }
                }
            ],
            "cause": [
                {
                    "stacktrace": [
                        {
                            "vars": {
                                "foo": "bar",
                                "password": "hello",
                                "secret": "hello",
                                "sensitive-stacktrace-val": "bye",
                            }
                        }
                    ],
                    "cause": [
                        {
                            "stacktrace": [
                                {
                                    "vars": {
                                        "foo": "bar",
                                        "password": "hello",
                                        "secret": "hello",
                                        "sensitive-stacktrace-val": "bye",
                                    }
                                }
                            ]
                        }
                    ],
                }
            ],
        }
    }

    result = processors.sanitize_stacktrace_locals(elasticapm_client, data)

    assert "stacktrace" in result["exception"]
    for stacktrace in (
        result["exception"]["stacktrace"],
        result["exception"]["cause"][0]["stacktrace"],
        result["exception"]["cause"][0]["cause"][0]["stacktrace"],
    ):
        assert len(stacktrace) == 1
        frame = stacktrace[0]
        assert "vars" in frame
        vars = frame["vars"]
        assert "foo" in vars
        assert vars["foo"] == "bar"
        assert "password" in vars
        assert vars["password"] == processors.MASK
        assert "secret" in vars
        assert vars["secret"] == processors.MASK
        assert "sensitive-stacktrace-val" in vars
        assert vars["sensitive-stacktrace-val"] == custom_field


def test_remove_http_request_body(http_test_data):
    assert "body" in http_test_data["context"]["request"]

    result = processors.remove_http_request_body(None, http_test_data)

    assert "body" not in result["context"]["request"]


@pytest.mark.parametrize(
    "elasticapm_client, custom_field, expected_header_cookies",
    [
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["custom-sensitive-cookie"]},
            {"custom-sensitive-cookie": processors.MASK},
            "foo=bar; password={0}; secret={0}; csrftoken={0}; custom-sensitive-cookie={0}".format(processors.MASK),
        ),
        (
            {},
            {"custom-sensitive-cookie": "123"},
            "foo=bar; password={0}; secret={0}; csrftoken={0}; custom-sensitive-cookie=123".format(processors.MASK),
        ),
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["custom-sensitive-*"]},
            {"custom-sensitive-cookie": processors.MASK},
            "foo=bar; password={0}; secret={0}; csrftoken={0}; custom-sensitive-cookie={0}".format(processors.MASK),
        ),
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["custom-sensitive"]},
            {"custom-sensitive-cookie": "123"},
            "foo=bar; password={0}; secret={0}; csrftoken={0}; custom-sensitive-cookie=123".format(processors.MASK),
        ),
    ],
    indirect=["elasticapm_client"],
)
def test_sanitize_http_request_cookies(elasticapm_client, custom_field, expected_header_cookies, http_test_data):
    http_test_data["context"]["request"]["headers"][
        "cookie"
    ] = "foo=bar; password=12345; secret=12345; csrftoken=abc; custom-sensitive-cookie=123"

    result = processors.sanitize_http_request_cookies(elasticapm_client, http_test_data)
    expected = {
        "foo": "bar",
        "password": processors.MASK,
        "secret": processors.MASK,
        "sessionid": processors.MASK,
    }

    expected.update(custom_field)

    assert result["context"]["request"]["cookies"] == expected

    assert result["context"]["request"]["headers"]["cookie"] == expected_header_cookies

    http_test_data["context"]["request"]["headers"].pop("cookie")
    http_test_data["context"]["request"]["headers"][
        "Cookie"
    ] = "foo=bar; password=12345; secret=12345; csrftoken=abc; custom-sensitive-cookie=123"

    result = processors.sanitize_http_request_cookies(elasticapm_client, http_test_data)
    expected = {
        "foo": "bar",
        "password": processors.MASK,
        "secret": processors.MASK,
        "sessionid": processors.MASK,
    }

    expected.update(custom_field)

    assert result["context"]["request"]["cookies"] == expected

    assert result["context"]["request"]["headers"]["Cookie"] == expected_header_cookies


@pytest.mark.parametrize(
    "elasticapm_client, expected_cookies",
    [
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["custom-sensitive-cookie"]},
            "foo=bar; httponly; secure ; sessionid={0}; httponly; secure; custom-sensitive-cookie={0}".format(
                processors.MASK
            ),
        ),
        (
            {},
            "foo=bar; httponly; secure ; sessionid={0}; httponly; secure; custom-sensitive-cookie=123".format(
                processors.MASK
            ),
        ),
    ],
    indirect=["elasticapm_client"],
)
def test_sanitize_http_response_cookies(elasticapm_client, expected_cookies, http_test_data):
    http_test_data["context"]["response"]["headers"][
        "set-cookie"
    ] = "foo=bar; httponly; secure ; sessionid=bar; httponly; secure; custom-sensitive-cookie=123"

    result = processors.sanitize_http_response_cookies(elasticapm_client, http_test_data)

    assert result["context"]["response"]["headers"]["set-cookie"] == expected_cookies


@pytest.mark.parametrize(
    "elasticapm_client, custom_header",
    [
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["some-header"]},
            {"some-header": processors.MASK},
        ),
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["some-*"]},
            {"some-header": processors.MASK},
        ),
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["other-val"]},
            {"some-header": "some-secret-value"},
        ),
        ({"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["*-*"]}, {"some-header": processors.MASK}),
    ],
    indirect=["elasticapm_client"],
)
def test_sanitize_http_headers(elasticapm_client, custom_header, http_test_data):
    result = processors.sanitize_http_headers(elasticapm_client, http_test_data)
    expected = {
        "cookie": "foo=bar; baz=foo",
        "foo": "bar",
        "password": processors.MASK,
        "secret": processors.MASK,
        "authorization": processors.MASK,
        "Ms-Client-Principal-Id": processors.MASK,
    }
    expected.update(custom_header)
    assert result["context"]["request"]["headers"] == expected
    assert result["context"]["response"]["headers"] == expected


@pytest.mark.parametrize(
    "elasticapm_client, custom_env",
    [
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["custom_env"]},
            {"custom_env": processors.MASK},
        ),
        ({}, {"custom_env": "bye"}),
    ],
    indirect=["elasticapm_client"],
)
def test_sanitize_http_wgi_env(elasticapm_client, custom_env, http_test_data):
    result = processors.sanitize_http_wsgi_env(elasticapm_client, http_test_data)
    expected = {
        "foo": "bar",
        "password": processors.MASK,
        "secret": processors.MASK,
    }
    expected.update(custom_env)
    assert result["context"]["request"]["env"] == expected


@pytest.mark.parametrize(
    "elasticapm_client, expected",
    [
        (
            {"sanitize_field_names": BASE_SANITIZE_FIELD_NAMES_UNPROCESSED + ["custom_field"]},
            "foo=bar&password={0}&secret={0}&cc=1234567890098765&custom_field={0}".format(processors.MASK),
        ),
        ({}, "foo=bar&password={0}&secret={0}&cc=1234567890098765&custom_field=123".format(processors.MASK)),
    ],
    indirect=["elasticapm_client"],
)
def test_post_as_string(elasticapm_client, expected, http_test_data):
    result = processors.sanitize_http_request_body(elasticapm_client, http_test_data)
    assert result["context"]["request"]["body"] == expected


def test_sanitize_dict():
    result = processors._sanitize("foo", {1: 2})
    assert result == {1: 2}


def test_non_utf8_encoding(elasticapm_client, http_test_data):
    broken = "broken=".encode("latin-1") + "aéöüa".encode("latin-1")
    http_test_data["context"]["request"]["headers"]["cookie"] = broken
    result = processors.sanitize_http_request_cookies(elasticapm_client, http_test_data)
    assert result["context"]["request"]["headers"]["cookie"] == "broken=a\ufffd\ufffd\ufffda"


def test_remove_stacktrace_locals():
    data = {"exception": {"stacktrace": [{"vars": {"foo": "bar", "password": "hello", "secret": "hello"}}]}}
    result = processors.remove_stacktrace_locals(None, data)

    assert "stacktrace" in result["exception"]
    stack = result["exception"]["stacktrace"]
    for frame in stack:
        assert "vars" not in frame


@processors.for_events(ERROR, TRANSACTION, SPAN)
def dummy_processor(client, data):
    data["processed"] = True
    return data


@processors.for_events()
def dummy_processor_no_events(client, data):
    data["processed_no_events"] = True
    return data


def dummy_processor_failing(client, data):
    raise ValueError("oops")


@pytest.mark.parametrize(
    "elasticapm_client",
    [{"processors": "tests.processors.tests.dummy_processor,tests.processors.tests.dummy_processor_no_events"}],
    indirect=True,
)
def test_transactions_processing(elasticapm_client):
    for i in range(5):
        elasticapm_client.begin_transaction("dummy")
        with elasticapm.capture_span("bla"):
            pass
        elasticapm_client.end_transaction("dummy_transaction", "success")
    for transaction in elasticapm_client.events[TRANSACTION]:
        assert transaction["processed"] is True
        assert "processed_no_events" not in transaction
    for span in elasticapm_client.events[SPAN]:
        assert span["processed"] is True
        assert "processed_no_events" not in span


@pytest.mark.parametrize(
    "elasticapm_client",
    [{"processors": "tests.processors.tests.dummy_processor,tests.processors.tests.dummy_processor_no_events"}],
    indirect=True,
)
def test_exception_processing(elasticapm_client):
    try:
        1 / 0
    except ZeroDivisionError:
        elasticapm_client.capture_exception()
    assert elasticapm_client.events[ERROR][0]["processed"] is True
    assert "processed_no_events" not in elasticapm_client.events[ERROR][0]


@pytest.mark.parametrize(
    "elasticapm_client",
    [{"processors": "tests.processors.tests.dummy_processor,tests.processors.tests.dummy_processor_no_events"}],
    indirect=True,
)
def test_message_processing(elasticapm_client):
    elasticapm_client.capture_message("foo")
    assert elasticapm_client.events[ERROR][0]["processed"] is True
    assert "processed_no_events" not in elasticapm_client.events[ERROR][0]


@mock.patch("elasticapm.base.constants.HARDCODED_PROCESSORS", ["tests.processors.tests.dummy_processor"])
@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "processors": "tests.processors.tests.dummy_processor,"
            "tests.processors.tests.dummy_processor_no_events,"
            "tests.processors.tests.dummy_processor"
        }
    ],
    indirect=True,
)
def test_deduplicate_processors(elasticapm_client):
    processors = elasticapm_client.load_processors()
    assert len(processors) == 2
    for p in processors:
        assert callable(p)


def test_for_events_decorator():
    @processors.for_events("error", "transaction")
    def foo(client, event):
        return True

    assert foo.event_types == {"error", "transaction"}


def test_drop_events_in_processor(elasticapm_client, caplog):
    dropping_processor = mock.MagicMock(return_value=None, event_types=[TRANSACTION], __name__="dropper")
    shouldnt_be_called_processor = mock.Mock(event_types=[])

    elasticapm_client._transport._processors = [dropping_processor, shouldnt_be_called_processor]
    with caplog.at_level(logging.DEBUG, logger="elasticapm.transport"):
        elasticapm_client.begin_transaction("test")
        elasticapm_client.end_transaction("test", "FAIL")
    assert dropping_processor.call_count == 1
    assert shouldnt_be_called_processor.call_count == 0
    assert elasticapm_client._transport.events[TRANSACTION][0] is None
    assert_any_record_contains(
        caplog.records, "Dropped event of type transaction due to processor mock.mock.dropper", "elasticapm.transport"
    )


@pytest.mark.parametrize(
    "elasticapm_client",
    [{"processors": "tests.processors.tests.dummy_processor,tests.processors.tests.dummy_processor_failing"}],
    indirect=True,
)
def test_drop_events_in_failing_processor(elasticapm_client, caplog):

    with caplog.at_level(logging.WARNING, logger="elasticapm.transport"):
        elasticapm_client.begin_transaction("test")
        elasticapm_client.end_transaction("test", "FAIL")
    assert elasticapm_client._transport.events[TRANSACTION][0] is None
    assert_any_record_contains(
        caplog.records,
        "Dropped event of type transaction due to exception in processor tests.processors.tests.dummy_processor_failing",
        "elasticapm.transport",
    )


def test_context_lines_processor(elasticapm_client):
    abs_path = os.path.join(os.path.dirname(__file__), "..", "utils", "stacks")
    fname1 = os.path.join(abs_path, "linenos.py")
    fname2 = os.path.join(abs_path, "linenos2.py")
    data = {
        "exception": {
            "stacktrace": [
                {"context_metadata": (fname1, 3, 2, None, None)},
                {"context_metadata": (fname2, 5, 2, None, None)},
                {"context_metadata": (fname1, 17, 2, None, None)},
                {"no": "context"},
            ]
        }
    }
    processed = processors.add_context_lines_to_frames(elasticapm_client, data)
    assert processed["exception"]["stacktrace"] == [
        {"pre_context": ["1", "2"], "context_line": "3", "post_context": ["4", "5"]},
        {"pre_context": ["c", "d"], "context_line": "e", "post_context": ["f", "g"]},
        {"pre_context": ["15", "16"], "context_line": "17", "post_context": ["18", "19"]},
        {"no": "context"},
    ]



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import pytest

from elasticapm.conf.constants import TRANSACTION
from elasticapm.traces import capture_span

memcache = pytest.importorskip("memcache")  # isort:skip


pytestmark = [pytest.mark.memcached]


if "MEMCACHED_HOST" not in os.environ:
    pytestmark.append(pytest.mark.skip("Skipping memcached tests, no MEMCACHED_HOST environment variable set"))


@pytest.mark.integrationtest
def test_memcached(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    host = os.environ.get("MEMCACHED_HOST", "localhost")
    with capture_span("test_memcached", "test"):
        conn = memcache.Client([host + ":11211"], debug=0)
        conn.set("mykey", "a")
        assert "a" == conn.get("mykey")
        assert {"mykey": "a"} == conn.get_multi(["mykey", "myotherkey"])
    elasticapm_client.end_transaction("BillingView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    expected_signatures = {"test_memcached", "Client.set", "Client.get", "Client.get_multi"}

    assert {t["name"] for t in spans} == expected_signatures

    assert spans[0]["name"] == "Client.set"
    assert spans[0]["type"] == "cache"
    assert spans[0]["subtype"] == "memcached"
    assert spans[0]["action"] == "query"
    assert spans[0]["parent_id"] == spans[3]["id"]

    assert spans[1]["name"] == "Client.get"
    assert spans[1]["type"] == "cache"
    assert spans[1]["subtype"] == "memcached"
    assert spans[1]["action"] == "query"
    assert spans[1]["parent_id"] == spans[3]["id"]
    assert spans[1]["context"]["destination"] == {
        "address": host,
        "port": 11211,
        "service": {"name": "", "resource": "memcached", "type": ""},
    }

    assert spans[2]["name"] == "Client.get_multi"
    assert spans[2]["type"] == "cache"
    assert spans[2]["subtype"] == "memcached"
    assert spans[2]["action"] == "query"
    assert spans[2]["parent_id"] == spans[3]["id"]

    assert spans[3]["name"] == "test_memcached"
    assert spans[3]["type"] == "test"

    assert len(spans) == 4


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": True,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "50ms",
        }
    ],
    indirect=True,
)
@pytest.mark.integrationtest
def test_memcache_span_compression(instrument, elasticapm_client):
    host = os.environ.get("MEMCACHED_HOST", "localhost")
    conn = memcache.Client([host + ":11211"], debug=0)
    elasticapm_client.begin_transaction("transaction.test")
    for i in range(5):
        conn.set(str(i), i)
    elasticapm_client.end_transaction("test")
    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert len(spans) == 1



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
import urllib.parse
from typing import TYPE_CHECKING, Optional, Tuple, Union

if TYPE_CHECKING:
    from asgiref.typing import ASGIApplication, ASGIReceiveCallable, ASGISendCallable, Scope, ASGISendEvent

import elasticapm
from elasticapm import Client, get_client, instrument
from elasticapm.conf import constants
from elasticapm.contrib.asyncio.traces import set_context
from elasticapm.utils import default_ports, encoding
from elasticapm.utils.disttracing import TraceParent


def wrap_send(send, middleware):
    @functools.wraps(send)
    async def wrapped_send(message) -> None:
        if message.get("type") == "http.response.start":
            await set_context(lambda: middleware.get_data_from_response(message, constants.TRANSACTION), "response")
            result = "HTTP {}xx".format(message["status"] // 100)
            elasticapm.set_transaction_result(result, override=False)
            elasticapm.set_transaction_outcome(http_status_code=message["status"], override=False)
        await send(message)

    return wrapped_send


class ASGITracingMiddleware:
    __slots__ = ("_app", "client")

    def __init__(self, app: "ASGIApplication", **options) -> None:
        self._app = app
        client = get_client()
        if not client:
            client = Client(**options)
        self.client = client
        if self.client.config.instrument and self.client.config.enabled:
            instrument()

    async def __call__(self, scope: "Scope", receive: "ASGIReceiveCallable", send: "ASGISendCallable") -> None:
        if scope["type"] != "http":
            await self._app(scope, receive, send)
            return
        send = wrap_send(send, self)
        wrapped_receive = receive
        url, url_dict = self.get_url(scope)
        body = None
        if not self.client.should_ignore_url(url):
            headers = self.get_headers(scope)
            self.client.begin_transaction(transaction_type="request", trace_parent=TraceParent.from_headers(headers))
            self.set_transaction_name(scope["method"], url)
            if scope["method"] in constants.HTTP_WITH_BODY and self.client.config.capture_body != "off":
                messages = []
                more_body = True
                while more_body:
                    message = await receive()
                    messages.append(message)
                    more_body = message.get("more_body", False)

                body_raw = b"".join([message.get("body", b"") for message in messages])
                body = str(body_raw, errors="ignore")

                # Dispatch to the ASGI callable
                async def new_wrapped_receive():
                    if messages:
                        return messages.pop(0)

                    # Once that's done we can just await any other messages.
                    return await receive()

                wrapped_receive = new_wrapped_receive

            await set_context(lambda: self.get_data_from_request(scope, constants.TRANSACTION, body), "request")

        try:
            await self._app(scope, wrapped_receive, send)
            elasticapm.set_transaction_outcome(constants.OUTCOME.SUCCESS, override=False)
            return
        except Exception as exc:
            self.client.capture_exception()
            elasticapm.set_transaction_result("HTTP 5xx", override=False)
            elasticapm.set_transaction_outcome(constants.OUTCOME.FAILURE, override=True)
            elasticapm.set_context({"status_code": 500}, "response")
            raise exc from None
        finally:
            self.client.end_transaction()

    def get_headers(self, scope_or_message: Union["Scope", "ASGISendEvent"]) -> dict[str, str]:
        headers = {}
        for k, v in scope_or_message.get("headers", {}):
            key = k.decode("latin1")
            val = v.decode("latin1")
            if key in headers:
                headers[key] = f"{headers[key]}, {val}"
            else:
                headers[key] = val
        return headers

    def get_url(self, scope: "Scope", host: Optional[str] = None) -> Tuple[str, dict[str, str]]:
        url_dict = {}
        scheme = scope.get("scheme", "http")
        server = scope.get("server", None)
        path = scope.get("root_path", "") + scope.get("path", "")

        url_dict["protocol"] = scheme + ":"

        if host:
            url = f"{scheme}://{host}{path}"
            url_dict["hostname"] = host
        elif server is not None:
            host, port = server
            url_dict["hostname"] = host
            if port:
                url_dict["port"] = port
            default_port = default_ports.get(scheme, None)
            if port != default_port:
                url = f"{scheme}://{host}:{port}{path}"
            else:
                url = f"{scheme}://{host}{path}"
        else:
            url = path
        qs = scope.get("query_string")
        if qs:
            query = "?" + urllib.parse.unquote(qs.decode("latin-1"))
            url += query
            url_dict["search"] = encoding.keyword_field(query)
        url_dict["full"] = encoding.keyword_field(url)
        return url, url_dict

    def get_ip(self, scope: "Scope", headers: dict) -> Optional[str]:
        x_forwarded_for = headers.get("x-forwarded-for")
        remote_addr = headers.get("remote-addr")
        ip: Optional[str] = None
        if x_forwarded_for:
            ip = x_forwarded_for.split(",")[0]
        elif remote_addr:
            ip = remote_addr
        elif scope.get("client"):
            ip = scope.get("client")[0]
        return ip

    async def get_data_from_request(self, scope: "Scope", event_type: str, body: Optional[str]) -> dict:
        """Loads data from incoming request for APM capturing.

        Args:
            request (Request)
            config (Config)
            event_type (str)
            body (str)

        Returns:
            dict
        """
        headers = self.get_headers(scope)
        result = {
            "method": scope["method"],
            "socket": {"remote_address": self.get_ip(scope, headers)},
            "cookies": headers.pop("cookies", {}),
        }
        if self.client.config.capture_headers:
            result["headers"] = headers
        if body and self.client.config.capture_body in ("all", event_type):
            result["body"] = body
        url, url_dict = self.get_url(scope)
        result["url"] = url_dict

        return result

    async def get_data_from_response(self, message: dict, event_type: str) -> dict:
        """Loads data from response for APM capturing.

        Args:
            message (dict)
            config (Config)
            event_type (str)

        Returns:
            dict
        """
        result = {}

        if "status" in message:
            result["status_code"] = message["status"]

        if self.client.config.capture_headers and "headers" in message:
            headers = self.get_headers(message)
            if headers:
                result["headers"] = headers

        return result

    def set_transaction_name(self, method: str, url: str) -> None:
        """
        Default implementation sets transaction name to "METHOD unknown route".
        Subclasses may add framework specific naming.
        """
        elasticapm.set_transaction_name(f"{method.upper()} unknown route")



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import binascii

import pytest

from elasticapm.utils.disttracing import TraceParent
from tests.utils import assert_any_record_contains


@pytest.mark.parametrize("tracing_bits,expected", [("00", {"recorded": 0}), ("01", {"recorded": 1})])
def test_tracing_options(tracing_bits, expected):
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-{}".format(tracing_bits)
    trace_parent = TraceParent.from_string(header)
    assert trace_parent.trace_options.recorded == expected["recorded"]


def test_unknown_header_components_ignored():
    header = "01-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03-xyz"
    trace_parent = TraceParent.from_string(header)
    assert trace_parent.to_string() == "01-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"


def test_trace_parent_to_str():
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    trace_parent = TraceParent.from_string(header)
    result = trace_parent.to_string()
    assert isinstance(result, str)
    assert header == result


def test_trace_parent_wrong_version(caplog):
    header = "xx-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    with caplog.at_level("DEBUG", "elasticapm.utils"):
        trace_parent = TraceParent.from_string(header)
    assert trace_parent is None
    assert_any_record_contains(caplog.records, "Invalid version field, value xx")


def test_trace_parent_wrong_version_255(caplog):
    """Version FF or 255 is explicitly forbidden"""
    header = "ff-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    with caplog.at_level("DEBUG", "elasticapm.utils"):
        trace_parent = TraceParent.from_string(header)
    assert trace_parent is None
    assert_any_record_contains(caplog.records, "Invalid version field, value 255")


def test_trace_parent_wrong_trace_options_field(caplog):
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-xx"
    with caplog.at_level("DEBUG", "elasticapm.utils"):
        trace_parent = TraceParent.from_string(header)
    assert trace_parent is None
    assert_any_record_contains(caplog.records, "Invalid trace-options field, value xx")


def test_trace_parent_wrong_format(caplog):
    header = "00"
    with caplog.at_level("DEBUG", "elasticapm.utils"):
        trace_parent = TraceParent.from_string(header)
    assert trace_parent is None
    assert_any_record_contains(caplog.records, "Invalid traceparent header format, value 00")


def test_trace_parent_binary():
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    tp = TraceParent.from_string(header)
    bytestr = tp.to_binary()
    print(type(bytestr[0]))
    assert binascii.hexlify(bytestr[0:1]) == b"00"  # version
    assert binascii.hexlify(bytestr[1:2]) == b"00"  # trace-id field identifier
    assert binascii.hexlify(bytestr[2:18]) == b"0af7651916cd43dd8448eb211c80319c"  # trace-id
    assert binascii.hexlify(bytestr[18:19]) == b"01"  # span-id field identifier
    assert binascii.hexlify(bytestr[19:27]) == b"b7ad6b7169203331"  # span-id
    assert binascii.hexlify(bytestr[27:28]) == b"02"  # trace-options field identifier
    assert binascii.hexlify(bytestr[28:29]) == b"03"  # trace-options

    tp2 = TraceParent.from_binary(bytestr)
    assert tp.version == tp2.version
    assert tp.trace_id == tp2.trace_id
    assert tp.span_id == tp2.span_id
    assert tp.trace_options == tp2.trace_options


def test_trace_parent_binary_invalid_length(caplog):
    with caplog.at_level("DEBUG", "elasticapm.utils"):
        tp = TraceParent.from_binary(b"123")
    assert tp is None
    assert_any_record_contains(
        caplog.records, "Invalid binary traceparent format, length is 3, should be 29, value b'123'"
    )


@pytest.mark.parametrize(
    "state_header",
    [
        "es=foo:bar;baz:qux,othervendor=<opaque>",
        "snes=x:y, es=foo:bar;baz:qux ,\tothervendor=<opaque>",  # whitespace (OWS)
        "othervendor=<opaque>,es=foo:bar;baz:qux",
    ],
)
def test_tracestate_parsing(state_header):
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    trace_parent = TraceParent.from_string(header, tracestate_string=state_header)
    assert trace_parent.tracestate == state_header
    assert trace_parent.tracestate_dict["foo"] == "bar"
    assert trace_parent.tracestate_dict["baz"] == "qux"
    assert len(trace_parent.tracestate_dict) == 2


@pytest.mark.parametrize("state_header", ["es=,othervendor=<opaque>", "foo=bar,baz=qux", ""])
def test_tracestate_parsing_empty(state_header):
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    trace_parent = TraceParent.from_string(header, tracestate_string=state_header)
    assert trace_parent.tracestate == state_header
    assert not trace_parent.tracestate_dict


def test_tracestate_adding_valid():
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    state_header = "es=foo:bar;baz:qux,othervendor=<opaque>"
    trace_parent = TraceParent.from_string(header, tracestate_string=state_header)
    trace_parent.add_tracestate("x", "y")
    assert trace_parent.tracestate_dict["x"] == "y"
    assert len(trace_parent.tracestate_dict) == 3
    trace_parent.add_tracestate("x", 1)
    assert trace_parent.tracestate_dict["x"] == "1"


@pytest.mark.parametrize("bad_char", ["\n", ":", ","])
def test_tracestate_adding_invalid(bad_char):
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    state_header = "es=foo:bar;baz:qux,othervendor=<opaque>"
    trace_parent = TraceParent.from_string(header, tracestate_string=state_header)
    trace_parent.add_tracestate("x", "y{}".format(bad_char))
    assert len(trace_parent.tracestate_dict) == 2
    assert "x" not in trace_parent.tracestate_dict
    trace_parent.add_tracestate("x{}".format(bad_char), "y")
    assert len(trace_parent.tracestate_dict) == 2
    assert "x{}".format(bad_char) not in trace_parent.tracestate_dict


def test_tracestate_length():
    header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    state_header = "es=foo:bar;baz:qux,othervendor=<opaque>"
    trace_parent = TraceParent.from_string(header, tracestate_string=state_header)
    trace_parent.add_tracestate("x", "y" * 256)
    assert len(trace_parent.tracestate_dict) == 2
    assert "x" not in trace_parent.tracestate_dict


def test_traceparent_from_header():
    header = "00-0000651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    legacy_header = "00-1111651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"
    tp1 = TraceParent.from_headers({"traceparent": header})
    tp2 = TraceParent.from_headers({"elastic-apm-traceparent": legacy_header})
    tp3 = TraceParent.from_headers({"traceparent": header, "elastic-apm-traceparent": legacy_header})
    assert tp1.to_string() == header
    assert tp2.to_string() == legacy_header
    # traceparent has precedence over elastic-apm-traceparent
    assert tp3.to_string() == header



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

pytest.importorskip("requests")  # isort:skip

import urllib.parse

import requests
from requests.exceptions import InvalidURL, MissingSchema

from elasticapm.conf import constants
from elasticapm.conf.constants import TRANSACTION
from elasticapm.traces import capture_span
from elasticapm.utils.disttracing import TraceParent

pytestmark = pytest.mark.requests


def test_requests_instrumentation(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        requests.get(url, allow_redirects=False)
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"].startswith("GET 127.0.0.1:")
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert url == spans[0]["context"]["http"]["url"]
    assert 200 == spans[0]["context"]["http"]["status_code"]
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % parsed_url.port,
        "type": "",
    }
    assert spans[0]["context"]["service"]["target"]["type"] == "http"
    assert spans[0]["context"]["service"]["target"]["name"] == f"127.0.0.1:{parsed_url.port}"
    assert spans[0]["outcome"] == "success"

    assert constants.TRACEPARENT_HEADER_NAME in waiting_httpserver.requests[0].headers
    trace_parent = TraceParent.from_string(
        waiting_httpserver.requests[0].headers[constants.TRACEPARENT_HEADER_NAME],
        tracestate_string=waiting_httpserver.requests[0].headers[constants.TRACESTATE_HEADER_NAME],
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    # this should be the span id of `requests`, not of urllib3
    assert trace_parent.span_id == spans[0]["id"]
    assert trace_parent.trace_options.recorded


@pytest.mark.parametrize("status_code", [400, 500])
def test_requests_instrumentation_error(instrument, elasticapm_client, waiting_httpserver, status_code):
    waiting_httpserver.serve_content("", code=status_code)
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        requests.get(url, allow_redirects=False)
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"].startswith("GET 127.0.0.1:")
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert url == spans[0]["context"]["http"]["url"]
    assert status_code == spans[0]["context"]["http"]["status_code"]
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % parsed_url.port,
        "type": "",
    }
    assert spans[0]["outcome"] == "failure"


def test_requests_instrumentation_via_session(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        s = requests.Session()
        s.get(url, allow_redirects=False)
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"].startswith("GET 127.0.0.1:")
    assert url == spans[0]["context"]["http"]["url"]

    assert constants.TRACEPARENT_HEADER_NAME in waiting_httpserver.requests[0].headers
    trace_parent = TraceParent.from_string(
        waiting_httpserver.requests[0].headers[constants.TRACEPARENT_HEADER_NAME],
        tracestate_string=waiting_httpserver.requests[0].headers[constants.TRACESTATE_HEADER_NAME],
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    # this should be the span id of `requests`, not of urllib3
    assert trace_parent.span_id == spans[0]["id"]
    assert trace_parent.trace_options.recorded


def test_requests_instrumentation_via_prepared_request(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        r = requests.Request("get", url)
        pr = r.prepare()
        s = requests.Session()
        s.send(pr, allow_redirects=False)
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"].startswith("GET 127.0.0.1:")
    assert url == spans[0]["context"]["http"]["url"]

    assert constants.TRACEPARENT_HEADER_NAME in waiting_httpserver.requests[0].headers
    trace_parent = TraceParent.from_string(
        waiting_httpserver.requests[0].headers[constants.TRACEPARENT_HEADER_NAME],
        tracestate_string=waiting_httpserver.requests[0].headers[constants.TRACESTATE_HEADER_NAME],
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    # this should be the span id of `requests`, not of urllib3
    assert trace_parent.span_id == spans[0]["id"]
    assert trace_parent.trace_options.recorded


def test_requests_instrumentation_malformed_none(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        with pytest.raises(MissingSchema):
            requests.get(None)


def test_requests_instrumentation_malformed_schema(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        with pytest.raises(MissingSchema):
            requests.get("")


def test_requests_instrumentation_malformed_path(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        with pytest.raises(InvalidURL):
            requests.get("http://")


def test_url_sanitization(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    url = url.replace("http://", "http://user:pass@")
    transaction_object = elasticapm_client.begin_transaction("transaction")
    requests.get(url)
    elasticapm_client.end_transaction("MyView")
    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]

    assert "pass" not in span["context"]["http"]["url"]
    assert constants.MASK_URL in span["context"]["http"]["url"]



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import socket
from functools import partial, partialmethod

import pytest

import elasticapm.utils
from elasticapm.conf import constants
from elasticapm.utils import (
    get_name_from_func,
    get_url_dict,
    getfqdn,
    nested_key,
    read_pem_file,
    sanitize_url,
    starmatch_to_regex,
    url_to_destination_resource,
)
from elasticapm.utils.deprecation import deprecated


@deprecated("alternative")
def deprecated_function():
    pass


def test_deprecation():
    with pytest.deprecated_call():
        deprecated_function()


@pytest.mark.parametrize(
    "url,expected",
    [
        (
            "http://example.com",
            {"protocol": "http:", "hostname": "example.com", "pathname": "", "full": "http://example.com"},
        ),
        (
            "http://example.com:443",
            {
                "protocol": "http:",
                "hostname": "example.com",
                "port": "443",
                "pathname": "",
                "full": "http://example.com:443",
            },
        ),
        (
            "http://example.com:443/a/b/c",
            {
                "protocol": "http:",
                "hostname": "example.com",
                "port": "443",
                "pathname": "/a/b/c",
                "full": "http://example.com:443/a/b/c",
            },
        ),
        (
            "https://example.com:443/",
            {
                "protocol": "https:",
                "hostname": "example.com",
                "port": "443",
                "pathname": "/",
                "full": "https://example.com:443/",
            },
        ),
        (
            "https://example.com:443/a/b/c?de",
            {
                "protocol": "https:",
                "hostname": "example.com",
                "port": "443",
                "pathname": "/a/b/c",
                "search": "?de",
                "full": "https://example.com:443/a/b/c?de",
            },
        ),
        (
            "https://[::ffff:a9fe:a9fe]/a/b/c?de",
            {
                "protocol": "https:",
                "hostname": "::ffff:a9fe:a9fe",
                "pathname": "/a/b/c",
                "search": "?de",
                "full": "https://[::ffff:a9fe:a9fe]/a/b/c?de",
            },
        ),
        (
            "http://[::ffff:a9fe:a9fe]:80/a/b/c?de",
            {
                "protocol": "http:",
                "hostname": "::ffff:a9fe:a9fe",
                "port": "80",
                "pathname": "/a/b/c",
                "search": "?de",
                "full": "http://[::ffff:a9fe:a9fe]:80/a/b/c?de",
            },
        ),
    ],
)
def test_get_url_dict(url, expected):
    assert get_url_dict(url) == expected


def test_get_name_from_func():
    def x():
        pass

    assert "tests.utils.tests.x" == get_name_from_func(x)


def test_get_name_from_func_class():
    class X(object):
        def x(self):
            pass

    assert "tests.utils.tests.x" == get_name_from_func(X.x)
    assert "tests.utils.tests.x" == get_name_from_func(X().x)


def test_get_name_from_func_partial():
    def x(x):
        pass

    p = partial(x, "x")
    assert "partial(tests.utils.tests.x)" == get_name_from_func(p)


def test_get_name_from_func_partialmethod_unbound():
    class X(object):
        def x(self, x):
            pass

        p = partialmethod(x, "x")

    assert "partial(tests.utils.tests.x)" == get_name_from_func(X.p)


def test_get_name_from_func_partialmethod_bound():
    class X(object):
        def x(self, x):
            pass

        p = partialmethod(x, "x")

    assert "partial(tests.utils.tests.x)" == get_name_from_func(X().p)


def test_get_name_from_func_lambda():
    assert "tests.utils.tests.<lambda>" == get_name_from_func(lambda x: "x")


@pytest.mark.parametrize(
    "pattern,input,match",
    [
        ("a*c", "abc", True),
        ("a*c", "abcd", False),
        ("a*c*", "abcd", True),
        ("a.c", "abc", False),
        ("a?c", "abc", False),
    ],
)
def test_starmatch_to_regex(pattern, input, match):
    re_pattern = starmatch_to_regex(pattern)
    assert bool(re_pattern.match(input)) is match, re_pattern.pattern


def test_url_sanitization():
    sanitized = sanitize_url("http://user:pass@localhost:123/foo?bar=baz#bazzinga")
    assert sanitized == "http://user:%s@localhost:123/foo?bar=baz#bazzinga" % constants.MASK_URL


def test_url_sanitization_urlencoded_password():
    sanitized = sanitize_url("http://user:%F0%9F%9A%B4@localhost:123/foo?bar=baz#bazzinga")
    assert sanitized == "http://user:%s@localhost:123/foo?bar=baz#bazzinga" % constants.MASK_URL


@pytest.mark.parametrize(
    "url,name,resource",
    [
        ("http://user:pass@testing.local:1234/path?query", "http://testing.local:1234", "testing.local:1234"),
        ("https://www.elastic.co:443/products/apm", "https://www.elastic.co", "www.elastic.co:443"),
        ("http://[::1]/", "http://[::1]", "[::1]:80"),
    ],
)
def test_url_to_destination(url, name, resource):
    destination = url_to_destination_resource(url)
    assert destination == resource


def test_url_to_destination_bad_port():
    destination = url_to_destination_resource("https://www.elastic.co:bad")
    assert destination == "www.elastic.co:443"


def test_read_pem_file():
    with open(os.path.join(os.path.dirname(__file__), "..", "ca", "ca.crt"), mode="rb") as f:
        result = read_pem_file(f)
        assert result.startswith(b"0\x82\x05{0\x82\x03c\xa0\x03\x02\x01\x02\x02\x14`c\xd8:\xe7")


def test_read_pem_file_chain():
    with open(os.path.join(os.path.dirname(__file__), "..", "ca", "chain.crt"), mode="rb") as f:
        result = read_pem_file(f)
        assert result.endswith(b"\xc8\xae")


@pytest.mark.parametrize(
    "data,key,expected",
    [
        (None, "x", None),
        ({}, "x", None),
        ({"x": 1}, "x", 1),
        ({"x": {"y": 1}}, "x.y", 1),
        ({"x": 1}, "x.y", None),
        ({"x": {"y": {}}}, "x.y.z", None),
    ],
)
def test_nested_key(data, key, expected):
    r = nested_key(data, *key.split("."))
    if expected is None:
        assert r is expected
    else:
        assert r == expected


def test_getfqdn(invalidate_fqdn_cache):
    assert getfqdn() == socket.getfqdn().lower()


def test_getfqdn_caches(invalidate_fqdn_cache):
    elasticapm.utils.fqdn = "foo"
    assert getfqdn() == "foo"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import pytest

from elasticapm.conf.constants import SPAN

psycopg = pytest.importorskip("psycopg")
pool_mod = pytest.importorskip("psycopg_pool")

pytestmark = [pytest.mark.psycopg_pool, pytest.mark.integrationtest]

has_postgres_configured = "POSTGRES_DB" in os.environ


def connect_kwargs():
    return {
        "dbname": os.environ.get("POSTGRES_DB", "elasticapm_test"),
        "user": os.environ.get("POSTGRES_USER", "postgres"),
        "password": os.environ.get("POSTGRES_PASSWORD", "postgres"),
        "host": os.environ.get("POSTGRES_HOST", None),
        "port": os.environ.get("POSTGRES_PORT", None),
    }


def make_conninfo():
    kw = connect_kwargs()
    host = kw["host"] or "localhost"
    port = kw["port"] or "5432"
    return f"postgresql://{kw['user']}:{kw['password']}@{host}:{port}/{kw['dbname']}"


@pytest.mark.skipif(not has_postgres_configured, reason="PostgreSQL not configured")
def test_connection_connect_generates_span(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("request")
    try:
        connection = psycopg.Connection.connect(**connect_kwargs())
        connection.close()
    finally:
        elasticapm_client.end_transaction("200")

    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 1
    assert spans[0]["action"] == "connect"


@pytest.mark.skipif(not has_postgres_configured, reason="PostgreSQL not configured")
def test_pool_generates_query_span(instrument, elasticapm_client):
    with pool_mod.ConnectionPool(
        make_conninfo(),
        min_size=1,
        max_size=2,
    ) as pool:
        pool.wait()

        elasticapm_client.begin_transaction("request")
        try:
            with pool.connection() as connection:
                with connection.cursor() as cursor:
                    cursor.execute("SELECT 1")
                    cursor.fetchone()
        finally:
            elasticapm_client.end_transaction("200")

    spans = elasticapm_client.events[SPAN]
    span = next(span for span in spans if span["action"] == "query")
    assert span["action"] == "query"
    assert span["context"]["db"]["type"] == "sql"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import urllib.parse
from collections import namedtuple

from elasticapm.conf import constants
from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import SpanType, capture_span, execution_context
from elasticapm.utils.disttracing import TraceParent
from elasticapm.utils.logging import get_logger

logger = get_logger("elasticapm.instrument")

SQS_MAX_ATTRIBUTES = 10


HandlerInfo = namedtuple("HandlerInfo", ("signature", "span_type", "span_subtype", "span_action", "context"))

# Used for boto3 < 1.7
endpoint_to_service_id = {"SNS": "SNS", "S3": "S3", "DYNAMODB": "DynamoDB", "SQS": "SQS"}


class BotocoreInstrumentation(AbstractInstrumentedModule):
    name = "botocore"

    instrument_list = [("botocore.client", "BaseClient._make_api_call")]

    capture_span_ctx = capture_span

    def _call(self, service, instance, args, kwargs):
        """
        This is split out from `call()` so that it can be re-used by the
        aiobotocore instrumentation without duplicating all of this code.
        """
        operation_name = kwargs.get("operation_name", args[0])

        parsed_url = urllib.parse.urlparse(instance.meta.endpoint_url)
        context = {
            "destination": {
                "address": parsed_url.hostname,
                "port": parsed_url.port,
                "cloud": {"region": instance.meta.region_name},
            }
        }

        handler_info = None
        handler = handlers.get(service, False)
        if handler:
            handler_info = handler(operation_name, service, instance, args, kwargs, context)
        if not handler_info:
            handler_info = handle_default(operation_name, service, instance, args, kwargs, context)

        return self.capture_span_ctx(
            handler_info.signature,
            span_type=handler_info.span_type,
            leaf=True,
            span_subtype=handler_info.span_subtype,
            span_action=handler_info.span_action,
            extra=handler_info.context,
        )

    def _get_service(self, instance):
        service_model = instance.meta.service_model
        if hasattr(service_model, "service_id"):  # added in boto3 1.7
            service = service_model.service_id
        else:
            service = service_model.service_name.upper()
            service = endpoint_to_service_id.get(service, service)
        return service

    def call(self, module, method, wrapped, instance, args, kwargs):
        service = self._get_service(instance)

        ctx = self._call(service, instance, args, kwargs)
        with ctx as span:
            if service in pre_span_modifiers:
                pre_span_modifiers[service](span, args, kwargs)
            result = wrapped(*args, **kwargs)
            if service in post_span_modifiers:
                post_span_modifiers[service](span, args, kwargs, result)
            request_id = result.get("ResponseMetadata", {}).get("RequestId")
            if request_id:
                span.update_context("http", {"request": {"id": request_id}})
            return result


def handle_s3(operation_name, service, instance, args, kwargs, context):
    span_type = "storage"
    span_subtype = "s3"
    span_action = operation_name
    if len(args) > 1:
        bucket = args[1].get("Bucket", "")
        key = args[1].get("Key", "")
    else:
        # TODO handle Access Points
        bucket = ""
        key = ""
    if bucket or key:
        context["otel_attributes"] = {}
        if bucket:
            context["otel_attributes"]["aws.s3.bucket"] = bucket
        if key:
            context["otel_attributes"]["aws.s3.key"] = key
    signature = f"S3 {operation_name} {bucket}"

    context["destination"]["service"] = {"name": span_subtype, "resource": f"s3/{bucket}", "type": span_type}

    return HandlerInfo(signature, span_type, span_subtype, span_action, context)


def handle_dynamodb(operation_name, service, instance, args, kwargs, context):
    span_type = "db"
    span_subtype = "dynamodb"
    span_action = "query"
    if len(args) > 1 and "TableName" in args[1]:
        table = args[1]["TableName"]
    else:
        table = ""
    signature = f"DynamoDB {operation_name} {table}".rstrip()

    context["db"] = {"type": "dynamodb", "instance": instance.meta.region_name}
    if operation_name == "Query" and len(args) > 1 and "KeyConditionExpression" in args[1]:
        context["db"]["statement"] = args[1]["KeyConditionExpression"]

    context["destination"]["service"] = {"name": span_subtype, "resource": table, "type": span_type}
    return HandlerInfo(signature, span_type, span_subtype, span_action, context)


def handle_sns(operation_name, service, instance, args, kwargs, context):
    if operation_name != "Publish":
        # only "publish" is handled specifically, other endpoints get the default treatment
        return False
    span_type = "messaging"
    span_subtype = "sns"
    span_action = "send"
    topic_name = ""
    if len(args) > 1:
        if "Name" in args[1]:
            topic_name = args[1]["Name"]
        if "TopicArn" in args[1]:
            topic_name = args[1]["TopicArn"].rsplit(":", maxsplit=1)[-1]
    signature = f"SNS {operation_name} {topic_name}".rstrip()
    context["destination"]["service"] = {
        "name": span_subtype,
        "resource": f"{span_subtype}/{topic_name}" if topic_name else span_subtype,
        "type": span_type,
    }
    return HandlerInfo(signature, span_type, span_subtype, span_action, context)


SQS_OPERATIONS = {
    "SendMessage": {"span_action": "send", "signature": "SEND to"},
    "SendMessageBatch": {"span_action": "send_batch", "signature": "SEND_BATCH to"},
    "ReceiveMessage": {"span_action": "receive", "signature": "RECEIVE from"},
    "DeleteMessage": {"span_action": "delete", "signature": "DELETE from"},
    "DeleteMessageBatch": {"span_action": "delete_batch", "signature": "DELETE_BATCH from"},
}


def handle_sqs(operation_name, service, instance, args, kwargs, context):
    op = SQS_OPERATIONS.get(operation_name, None)
    if not op:
        # only "publish" is handled specifically, other endpoints get the default treatment
        return False
    span_type = "messaging"
    span_subtype = "sqs"
    topic_name = ""

    if len(args) > 1:
        topic_name = args[1]["QueueUrl"].rsplit("/", maxsplit=1)[-1]
    signature = f"SQS {op['signature']} {topic_name}".rstrip() if topic_name else f"SQS {op['signature']}"
    context["destination"]["service"] = {
        "name": span_subtype,
        "resource": f"{span_subtype}/{topic_name}" if topic_name else span_subtype,
        "type": span_type,
    }
    return HandlerInfo(signature, span_type, span_subtype, op["span_action"], context)


def modify_span_sqs_pre(span, args, kwargs) -> None:
    operation_name = kwargs.get("operation_name", args[0])
    if span.id:
        trace_parent = span.transaction.trace_parent.copy_from(span_id=span.id)
    else:
        # this is a dropped span, use transaction id instead
        transaction = execution_context.get_transaction()
        trace_parent = transaction.trace_parent.copy_from(span_id=transaction.id)
    attributes = {constants.TRACEPARENT_HEADER_NAME: {"DataType": "String", "StringValue": trace_parent.to_string()}}
    if trace_parent.tracestate:
        attributes[constants.TRACESTATE_HEADER_NAME] = {"DataType": "String", "StringValue": trace_parent.tracestate}
    if len(args) > 1:
        if operation_name in ("SendMessage", "SendMessageBatch"):
            attributes_count = len(attributes)
            if operation_name == "SendMessage":
                messages = [args[1]]
            else:
                messages = args[1]["Entries"]
            for message in messages:
                message["MessageAttributes"] = message.get("MessageAttributes") or {}
                if len(message["MessageAttributes"]) + attributes_count <= SQS_MAX_ATTRIBUTES:
                    message["MessageAttributes"].update(attributes)
                else:
                    logger.info("Not adding disttracing headers to message due to attribute limit reached")
        elif operation_name == "ReceiveMessage":
            message_attributes = args[1].setdefault("MessageAttributeNames", [])
            if "All" not in message_attributes:
                message_attributes.extend([constants.TRACEPARENT_HEADER_NAME, constants.TRACESTATE_HEADER_NAME])


def modify_span_sqs_post(span: SpanType, args, kwargs, result) -> None:
    operation_name = kwargs.get("operation_name", args[0])
    if operation_name == "ReceiveMessage" and "Messages" in result:
        for message in result["Messages"][:1000]:  # only up to 1000 span links are recorded
            if "MessageAttributes" in message and constants.TRACEPARENT_HEADER_NAME in message["MessageAttributes"]:
                tp = TraceParent.from_string(
                    message["MessageAttributes"][constants.TRACEPARENT_HEADER_NAME]["StringValue"]
                )
                span.add_link(tp)


def handle_default(operation_name, service, instance, args, kwargs, context):
    span_type = "aws"
    span_subtype = service.lower()
    span_action = operation_name

    context["destination"]["service"] = {"name": span_subtype, "resource": span_subtype, "type": span_type}

    signature = f"{service}:{operation_name}"
    return HandlerInfo(signature, span_type, span_subtype, span_action, context)


handlers = {
    "S3": handle_s3,
    "DynamoDB": handle_dynamodb,
    "SNS": handle_sns,
    "SQS": handle_sqs,
    "default": handle_default,
}

pre_span_modifiers = {
    "SQS": modify_span_sqs_pre,
}

post_span_modifiers = {
    "SQS": modify_span_sqs_post,
}
